package WordNet;

use 5.006;
use strict;
use warnings;

my (@vars, @util);
BEGIN {
	@vars = qw/@pos @pos_extended %pos_num %pos_short %pos_medium %pos_long
            @pointers %english %querydata %ptr_hash %ptr_short %ptr_sym
            @ptr_order %ptr_order/;
}

my $hi_res;
BEGIN {
	if(eval "require Time::HiRes; 1") {
		Time::HiRes->import(qw/time/);
		$hi_res = 1;
	} else {
		$hi_res = 0;
	}
}
my $program_start = time;
my $term_supports_title = (($ENV{TERM}||'') =~ /^(?:xterm|linux)$/) ? 1 : 0;
use Getopt::Long;
BEGIN { push @util, qw/&counter/; } sub counter {
	shift if @_ and ref $_[0];
	my $options = Getopt::Long::Parser->new;
	$options->configure(pass_through => 'prefix_pattern=|-');
	local @ARGV = @_;
	my $display = '';
	$options->getoptions(
		'display|counter=s' => \$display,
		'title|termtitle|term!' => \ (my $title = $term_supports_title),
		'mod|every=i' => \ (my $mod = 0),
		'wait|time=f' => \ (my $wait = 5),
		'persecond' => \ (my $persec = 0),
		'<>' => sub { $display = shift; },
	);
	my $c = 0;
	my $t = time;
	my $nt = $t;
	my $fmt = $hi_res ? '%5.2f' : '%3d';
	return sub {
		if ($display and (@_ or ($mod ? (!($c % $mod)) : (time > $nt)))) {
			my $time = time;
			warn sprintf "%s%s %6d -- %4d (total: $fmt)\n",
				($title ? "\e]2;$display $c\007" : ""),
				$display, $c, int($time-$t), ($time-$program_start);
			$nt = $t + $wait * (1 + int(($time - $t)/$wait)) unless $mod;
		}
		++$c;
	};
}

BEGIN {
    use vars @vars;
    @pos = qw/n v a r/;
    @pos_extended = (@pos, qw/s/);
    %pos_short = qw/noun n verb v adj a adjective a adv r adverb r
        sat s satellite s n n v v a a r r s s 1 n 2 v 3 a 4 r 5 s/;
    %pos_medium = qw/n noun v verb a adj r adv s sat/;
    %pos_long = qw/n noun v verb a adjective r adverb s satellite/;
    for my $i (0..$#pos_extended) {
        $pos_num{$i + 1} = $i + 1;
        for my $h (\%pos_short, \%pos_medium, \%pos_long) {
            $pos_num{$h->{$pos_extended[$i]}} = $i + 1;
        }
    }
    @pointers =
	#   short,  sym,   regex,  is_fake, full name,                     WN::QD,
        ( '@',  '@',   qr/\@/,       0, 'hypernym',                    'hype',
          'I',  '@i',  qr/\@i/,      0, 'instance',                    'inst',
          '',   '@i?', qr/\@i?/,     1, 'hypernym/instance',           'hypes',
          '~',  '~',   qr/~/,        0, 'hyponym',                     'hypo',
          'i',  '~i',  qr/~i/,       0, 'has instance',                'hasi',
          '',   '~i?', qr/~i?/,      1, 'hyponym/has instance',        'hypos',
          '!',  '!',   qr/!/,        0, 'antonym',                     'ants',
          'm',  '%m',  qr/[%]m/,     0, 'member meronym',              'mmem|mmer',
          'p',  '%p',  qr/[%]p/,     0, 'part meronym',                'mprt|pmer',
          's',  '%s',  qr/[%]s/,     0, 'substance meronym',           'msub|smer',
          '',   '%',   qr/[%][mps]/, 1, 'all meronyms',                'mero',
          'M',  '#m',  qr/\#m/,      0, 'member holonym',              'hmem|mhol',
          'P',  '#p',  qr/\#p/,      0, 'part holonym',                'hprt|phol',
          'S',  '#s',  qr/\#s/,      0, 'substance holonym',           'hsub|shol',
          '',   '#',   qr/\#[mps]/,  1, 'all holonyms',                'holo',
          '$',  '$',   qr/\$/,       0, 'verb group',                  'vgrp',
          '&',  '&',   qr/&/,        0, 'similar to',                  'sim',
          '*',  '*',   qr/\*/,       0, 'entailment',                  'enta',
          '+',  '+',   qr/\+/,       0, 'derivationally related form', 'deri',
          'C',  ';c',  qr/;c/,       0, 'domain category',             'dmnc',
          'R',  ';r',  qr/;r/,       0, 'domain region',               'dmnr',
          'U',  ';u',  qr/;u/,       0, 'domain usage',                'dmnu',
          '',   ';',   qr/;[cru]/,   1, 'domain all',                  'domn',
          'c',  '-c',  qr/-c/,       0, 'domain term category',        'dmtc',
          'r',  '-r',  qr/-r/,       0, 'domain term region',          'dmtr',
          'u',  '-u',  qr/-u/,       0, 'domain term usage',           'dmtu',
          '',   '-',   qr/-[cru]/,   1, 'domain term all',             'domt',
          '<',  '<',   qr/</,        0, 'participle',                  'part',
          '=',  '=',   qr/=/,        0, 'attribute',                   'attr',
          '>',  '>',   qr/>/,        0, 'cause',                       'caus',
          '\\', '\\',  qr/\\/,       0, 'pertainym',                   'pert',
          '^',  '^',   qr/\^/,       0, 'see also',                    'also',
          '',   '.*',  qr/.*/,       1, 'any pointer',                 'any',
          '',   '.*',  qr/.*/,       1, 'all pointers',                'all',
		  );
	my @maps =
		( \%WordNet::ptr_regex,     undef,
	      \%WordNet::is_pseudo_ptr, undef,
	      \%english,                \%WordNet::english_internal,
	      \%querydata,              \%WordNet::querydata_internal,
	      );
	my $per_ptr = 2 + (@maps / 2);
	for my $i (0..$#pointers/$per_ptr) {
		my ($short, $sym, @other) =
			@pointers[($i*$per_ptr)..(($i+1)*$per_ptr - 1)];
		$ptr_short{$sym} = $short if $short;
		$ptr_sym{$_} = $sym for grep $_, $short, $sym;
		for my $j (0..$#other) {
			my @o = split /\|/, $other[$j];
			my ($sym2o, $o2sym) = @maps[2*$j, 2*$j+1];
			if ($sym2o) { $sym2o->{$sym} = $_ for @o; }
			if ($o2sym) { $o2sym->{$_} = $sym for @o; }
        }
        push @ptr_order, $sym unless $WordNet::is_pseudo_ptr{$sym};
    }
    %ptr_order = map { $ptr_order[$_] => $_ } 0..$#ptr_order;
	for my $ptr (keys %WordNet::ptr_regex) {
		$WordNet::ptr_array{$ptr} =
			[ grep /$WordNet::ptr_regex{$ptr}/, @ptr_order ];
		$ptr_hash{$ptr} =
			{ map { ($_=>1) } @{$WordNet::ptr_array{$ptr}} };
	}
	for my $intern (@maps[grep { ($_ % 2) == 1 } 0..$#maps]) {
		next unless $intern;
		$intern->{$_} = $ptr_hash{$intern->{$_}} for keys %$intern;
		$ptr_hash{$_} = $intern->{$_} for keys %$intern;
	}
}

# use Memoize; # caused weirdness with references
sub __pointermap {
	# possible inputs:
	# string (QueryData style, English, pointer symbol, extended ptr symbol)
	# string|string (combo of above string types)
	# [string|string] (array ref)
	# { sym => 1, sym => 1 } (hash ref -- keys MUST be actual pointer symbols)
	# regex (will be matched against pointer symbols)
	my $type = shift;
	my $ref = ref $type;
	$type = $ptr_sym{$type} if !$ref and exists $ptr_sym{$type};
	return $type if $ref eq 'HASH' or $ref eq 'Regexp';
	if ($ref eq 'ARRAY' or (!$ref and $type =~ /\|/)) {
		return +{(map %$_, map __pointermap($_), $ref?(@$type):(split /\|/, $type))};
	}
	return ((exists $ptr_hash{$type})
	        ? $ptr_hash{$type}
			: qr/\Q$type\E/) unless $ref;
	return +{};
}
# memoize('__pointermap', INSTALL=>'_pointermap');
*_pointermap = \&__pointermap; # needed since I dropped Memoize

require Exporter;

our @ISA = qw(Exporter);
our %EXPORT_TAGS = ( 'all'  => [ @vars ],
                     'vars' => [ @vars ],
                     'pos'  => [ @vars ] );
our @EXPORT_OK   = (@vars, @util, '$wn', '$i');
our @EXPORT      = ();
our $VERSION = '0.01';

our ($wn, $i);
sub import {
	my ($make_wn, $make_i, $disp) = (0, 0, '');
	@_ = grep { not
		  /^:?wn$/i                        ? (          $make_wn = 1)
		: /^(?::?wn)?i(?:ter(?:ator)?)?(?:(?:d(?:isp)?)?:(.*))?$/i
		? (($make_i, $make_wn, $disp) = (1, 1, $1))
		: 0
	} @_;
    $make_wn and ($wn or $wn = __PACKAGE__->new);
	$make_i  and ($i  or $i  = $wn->iterator(grep $disp, display => $disp));
    __PACKAGE__->export_to_level(1, $_[0], '$wn') if $wn;
    __PACKAGE__->export_to_level(1, $_[0], '$i')  if $i;
	goto &Exporter::import;
}

our @defaults = ( reader    =>     'Disk',
                  writer    =>     'Disk',
                  read_only =>          1,
                  case_sensitive =>     0,
                  retain_underscore =>  0,
                  conflate35 =>         1 );

sub new {
    my $class = shift;
    my $ref = { @defaults, @_ };
    my $self = bless $ref, $class;
    $ref->{WN} = $self;

    my $readclass = "$class\::IO\::$ref->{reader}";
    eval "require $readclass;"
        or die "In $class\->new: Error loading '$readclass': $@";
    $ref->{READER} = $readclass->new(%$ref);

    unless ($ref->{read_only}) {
        my $writeclass = "$class\::IO\::$ref->{writer}";
        eval "require $writeclass;"
            or die "In $class\->new: Error loading '$writeclass': $@";
        $ref->{WRITER} = $writeclass->new(%$ref);
    }
    return $self;
}

sub fetch {
    my $self = shift;
    my ($word, @rest) = @_;
    $word = lc $word unless $self->{case_sensitive};
    $word =~ tr/_/ / unless $self->{retain_underscore};
    my (@ps, @senses) = (); # MODPERL requires initialization
    while (@rest and defined $pos_short{$rest[0]}) { push @ps, shift @rest; }
    @ps = map { $_ eq 's' ? 'a' : $_; } @ps if $self->{conflate35};
    @senses = @rest;
    @ps = @pos if not @ps;
    @senses = (0) if not @senses;

    my @ret = (); # MODPERL requires initialization
    foreach my $pos (@ps) {
        foreach my $sense (@senses) {
            foreach my $id ($self->{READER}->id($word, $pos, $sense)) {
                my $ss = $self->{READER}->read($id);
                if ($self->{case_sensitive}) {
                    foreach my $w ($ss->words) {
                        if ($w eq $word) {
                            push @ret, $ss;
                            last;
                        }
                    }
                } else {
                    push @ret, $ss;
                }
            }
        }
    }
    return wantarray ? @ret : ((@ret) ? $ret[0] : undef);
}

sub fetch_by_sk {
    my $self = shift;
    my $id = $self->{READER}->id_by_sk(@_);
    return unless $id;
    return $self->{READER}->read($id);
}

sub fetch_by_qd {
    my $self = shift;
    my $id = $self->{READER}->id_by_qd(@_);
    return unless $id;
    return $self->{READER}->read($id);
}

sub fetch_by_id {
    my $self = shift;
    my $id = shift;
    $self->{READER}->read($id);
}

sub pretty { shift->sprint . "\n"; }
sub sprint {
    my $self = shift;
    my ($word) = @_;
    $word = lc $word unless $self->{case_sensitive};
    $word =~ tr/_/ / unless $self->{retain_underscore};
    my $text = '';
    foreach my $pos (@pos) {
        my @senses = $self->fetch($word, $pos);
        next unless @senses;
        $text .= "The $pos_long{$pos} $word has " . scalar(@senses) . ' sense' . ((@senses > 1)?'s':'') . "\n\n";
        my $sns = 0;
        while (@senses) {
            $sns++;
            $text .= "$sns. " . $senses[0]->pretty;
            shift @senses;
        }
    }
    return $text;
}

# Handled by READER
sub iterator {          shift->{READER}->iterator(@_); }
sub is_word {           shift->{READER}->is_word(@_);  }
sub search {            shift->{READER}->search(@_);   }
sub id_by_sk {          shift->{READER}->id_by_sk(@_); }
sub sk_by_id {          shift->{READER}->sk_by_id(@_); }
sub sk_to_sn {          shift->{READER}->sk_to_sn(@_); }
sub sk_to_sensenumber { shift->{READER}->sk_to_sn(@_); }
sub _follow_pointer {   shift->{READER}->read(@_);     }
sub filenames {         shift->{READER}->filenames(@_) }

# Handled by WRITER
sub store {             shift->{WRITER}->write(@_);    }
sub delete {            shift->{WRITER}->delete(@_);   }

package WordNet::Type;
use overload '""' => 'sprint', 'cmp' => 'strcmp', fallback => 1;
sub strcmp { "$_[0]" cmp "$_[1]" }

sub new { my $class=shift; bless { id => '', pos => '',
                                   @_, class=>$class }, $class; }

sub pos { $_[0]->{pos}; }
sub id  { $_[0]->{id};  }

package WordNet::Common;

our @ISA = qw/WordNet::Type/;

sub _wn_reader {
    my $s = shift;
    my $w = $s->{WN};
    my $h = $w->{READER};
    $h = $h->{cacheREADER} if $w->{reader} eq 'Cache';
    $h;
}

sub new { shift->SUPER::new( pointers => {}, frames => [], @_ ); }

sub pointers {
    my $self = shift;
    my ($type, $number) = @_;
    $number = 0 unless defined $number and int $number;
    return wantarray ? () : undef if not defined $self->{pointers};
    my @keys = keys %{$self->{pointers}};
    if ($type) {
        my $ref = ref $type;
        $type = WordNet::_pointermap($type);
        if ('HASH' eq ref $type) {
            @keys = grep $type->{$_}, @keys;
        } else {
            @keys = grep /$type/, @keys;
        }
    }
    my @ret = map { (@{$self->{pointers}{$_}}) } @keys;
    if ($number) {
        return wantarray ? () : 0 if $number > @ret;
        $number++ if $number < 0;
        return $ret[$number - 1];
    }
    @ret;
}

sub frames {
    my $self = shift;
    return wantarray ? () : undef if not defined $self->{frames};
    @{$self->{frames}};
}

sub framestext {
    my $self = shift;
    my $r = $self->_wn_reader;
    map { $r->{frames}{$_} } $self->frames;
}

package WordNet::Synset;

our @ISA = qw/WordNet::Common/;

sub lexfilename {
    my $s = shift;
    $s->_wn_reader->{lexfilename}{$s->{lexfile}};
}

sub new { shift->SUPER::new( words => [], gloss => {}, @_ ); }

sub all_pointers {
    my $self = shift;
    map $_->pointers(@_), $self, $self->Words;
}

sub words {
    my $self = shift;
    return unless defined wantarray;
    return wantarray ? () : undef if not defined $self->{words};
    map $_->word, @{$self->{words}};
}

sub Words {
    my $self = shift;
    return unless defined wantarray;
    return wantarray ? () : undef if not defined $self->{words};
    @{$self->{words}};
}

sub gloss {
    my $self = shift;
    my $type = shift || 'original';
    return undef if not defined $self->{gloss};
    $self->{gloss}{$type};
}

sub pretty { shift->sprint . "\n"; }
sub sprint {
    my $self = shift;
    my @w = $self->words;
    return '' unless @w;
    my $string = join ', ', @w;
    $string .= ' -- (' . $self->gloss . ')';
}

sub wordnet_form { # differs from tagging_form only in the tr///'s
    my $self = shift;
    my ($word, $object_oriented) = @_;
    $word = ($self->words)[0] unless defined $word;
    $word =~ tr/A-Z /a-z_/;
    foreach my $W (@{$self->{words}}) {
        my $w = $W->word;
        $w =~ tr/ /_/; # WordNet form has '_' not ' '
        (my $test = $w) =~ tr/A-Z /a-z_/;
        return(($object_oriented) ? $W : $w) if $test eq $word;
    }
    my $ret = ($self->Words)[0];
    return $object_oriented ? $ret : $ret->word;
}

sub tagging_form {
    my $self = shift;
    my ($word, $object_oriented) = @_;
    $word = ($self->words)[0] unless defined $word;
    $word =~ tr/-_ //d; # BUG? should be tr/ /_/
    foreach my $W (@{$self->{words}}) {
        my $w = $W->word;
        $w =~ tr/ /_/; # WordNet form has '_' not ' '
        my $test = $w;
        $test =~ tr/-_ //d; # BUG? should be tr/ /_/
        return(($object_oriented) ? $W : $w) if (lc $test eq lc $word);
    }
    my $ret = ($self->Words)[0];
    return $object_oriented ? $ret : $ret->word;
}

my %convert_back = qw/n 1 v 2 a 3 r 4 s 5/;
sub sk {
    my $self = shift;
    my ($word) = @_;
    $word = $self->{words}[0]->word unless defined $word;
    my $headword = $self->wordnet_form($word, 1);
    my $sk = $headword->word;
    $sk =~ tr/ /_/;
    $sk = lc $sk . '%';
    $sk .= $convert_back{$headword->pos} . ':';
    $sk .= sprintf('%02d', $self->{lexfile} || 0) . ':';
    $sk .= sprintf('%02d', $headword->{lexsense} || 0) . ':';
    return "$sk:" unless $self->{pos} =~ /[as]/;
    return "$sk:" if $self->{words}[0]->pointers('!');
    my @ptrs = $self->pointers('&', 1);
    return "$sk:" unless @ptrs;
    my $headss = $ptrs[0]->follow;
    return "$sk:" unless $headss;
    $headword = $headss->wordnet_form(undef, 1);
    $sk .= $headss->wordnet_form . ':';
    $sk .= sprintf('%02d', $headword->{lexsense});
    $sk;
}

sub querydata_id {
    my $self = shift;
    my ($word) = @_;
    $word = ($self->words)[0] unless defined $word;
    $word = $self->wordnet_form($word);
    my ($Word) = grep { $self->wordnet_form($_->word) eq $word } $self->Words;
    my $sense = $Word->sensenumber;
    my $pos = $self->pos;
    $pos = 'a' if $pos eq 's';
    return "$word#$pos#$sense";
}
*qdid = \&querydata_id;

sub SS { $_[0] }

package WordNet::Word;
our @ISA = qw/WordNet::Common/;

sub new { shift->SUPER::new( word => '', lexsense => '', syntax => '', @_ ); }
sub sk { $_[0]->SS()->sk($_[0]->word) }

# Handled by {WN}{READER}
sub sensenumber { $_[0]->{WN}{READER}->sensenumber($_[0]->sk             ); }
*sense_number = \&sensenumber;
*sense_num = \&sensenumber;
*sensenum = \&sensenumber;
sub count       { $_[0]->{WN}{READER}->count(      $_[0]->sk             ); }
sub sentenceids { $_[0]->{WN}{READER}->sentenceids($_[0]->sk             ); }
sub sentences   { $_[0]->{WN}{READER}->sentences(  $_[0]->sk, $_[0]->word); }

sub lexfilename { $_[0]->SS()->lexfilename; }
sub word     { $_[0]->{word};     }
*words = \&word;
*pretty = \&word;
*sprint = \&word;
sub lexsense { $_[0]->{lexsense}; }
sub syntax   { $_[0]->{syntax};   }

sub SS {
    my $w = shift;
    return $w->{SS} if $w->{SS};
    $w->{WN}{READER}->parent_ss($w);
}

package WordNet::Pointer;
our @ISA = qw/WordNet::Type/;

sub new {
    shift->SUPER::new( type => '', toclass => '',
                       from => {}, to => {},
                       WN => undef, @_ );
}

sub type  { $_[0]->{type};    }
sub class { $_[0]->{toclass}; }
sub source { $_[0]->{from}{id}; } *src = \&source; *from = \&source;
sub target { $_[0]->{to}{id};   } *tar = \&target; *to   = \&target;
sub sprint { my $p = shift; join " ", map $p->$_, qw/source type target/; }
sub follow {
    my $self = shift;
    die "Failed to follow pointer [cannot determine associated WordNet object]\n" unless $self->{WN};
    $self->{WN}->_follow_pointer($self->{to}{id});
}

package WordNet::Iterator;
our @ISA = ();

sub new {
    my $class = shift;
    my $ref = { NEXT => \&next, RESET => \&reset, COUNT => \&count, @_ };
    $ref->{ORIGINAL} = $ref;
    bless $ref, $class;
}

sub count { $_[0]->{COUNT}(@_); }
sub next { $_[0]->{NEXT}(@_); }
sub reset {
    my $self = shift;
    if ($self->{RESET}) {
        $self->{RESET}->($self);
    } else {
        $self->{$_} = $self->{ORIGINAL}{$_} for keys %{$self->{ORIGINAL}};
    }
}

1;
__END__

=head1 NAME

WordNet - Perl Interface to Princeton WordNet

=head1 SYNOPSIS

  use WordNet;
  # new requires no options: these are the defaults
  my $wn = WordNet->new(reader =>        'Disk',
                        writer =>        'Disk',
                        read_only =>          0,
                        case_sensitive =>     0,
                        retain_underscore =>  0,
                        conflate35 =>         1 );
  # Iterate through every synset in WordNet
  my $i  = $wn->iterator;

  while (my $synset = $i->next) {
     print join(', ', $synset->words), "\n";
     print $synset->pretty;
     my @word_objects = $synset->Words;
     my @pointers = $synset->pointers; # no arguments => all SYNSET pointers
     my @allpointers = $synset->all_pointers;
     my @pointed_to = map { $_->follow } @pointers;
     for my $word ($synset->Words) {
         # can replace 'antonym' on next line with '!', qr/!/, or 'ants'
         # see top of file for list of relations and equivalent names
         if (my @antonyms = map { $_->follow } $word->pointers('antonym')) {
             print $word->word." is the opposite of ".$_->word."\n"
                 for @antonyms;
         }
     }
     print $synset->pos, " or ", $pos_medium{$synset->pos}, " or ",
           $pos_long{$synset->pos}, "\n";
     print "Synset ID: ", $synset->id, "\n";
     print "Better ID: ", $synset->sk, "\n";
     print "QueryData ID: ", $synset->qd_id, "\n";
     print "SKs change per word: ", $synset->sk($_), "\n" for $synset->words;
     print "So do QD ID's: ", $synset->qd_id($_),    "\n" for $synset->words;
  }



=head1 ABSTRACT

  Officially supported Perl interface to the WordNet database.

=head1 DESCRIPTION

  Officially supported Perl interface to WordNet.
  Currently, there are (much?) faster interfaces to the ASCII format WordNet files. This one will also offer access to the XML and SQL versions.

=head2 EXPORT

=over 4

=item :pos (alias :all, :vars)

BEGIN { @vars = qw/@pos @pos_extended %pos_num %pos_short %pos_medium %pos_long
            @pointers %english %querydata
            @ptr_order %ptr_order/; }
BEGIN {
    use vars @vars;
    @pos = qw/n v a r/;
    @pos_extended = (@pos, qw/s/);
    %pos_short = qw/noun n verb v adj a adjective a adv r adverb r
        sat s satellite s n n v v a a r r s s 1 n 2 v 3 a 4 r 5 s/;
    %pos_medium = qw/n noun v verb a adj r adv s sat/;
    %pos_long = qw/n noun v verb a adjective r adverb s satellite/;
C<@pos> -- Parts of speech as they should be passed to this module.
  qw(n v a r)

C<@pos_extended> -- Parts of speech including 's' for 'satellite'.
  qw(n v a r s)

C<%pos_medium> -- Parts of speech as used in filenames:
  n => 'noun', r => 'adv', ...

C<%pos_long> -- Parts of speech in English:
  n => 'noun', r => 'adverb', ...

=head2 METHODS

=over 4

=item new( [options] )

Creates and initializes a WordNet object, through which Synsets and Words can be C<fetch>ed and C<store>d.

Options:

=over 4

=item reader => $type, writer => $type

Specifies how the data will be read/written. I<WordNet::IO::$type.pm> needs to exist. Valid types are currently: 'Disk', 'Cache', 'XML', and 'NONE'. Default is: ( reader => 'Disk', writer => 'Disk' )

=item read_only => bool

If true, no writing is possible.
Default: false

=item case_sensitive => bool

If true, lookups are case-sensitive.
Default: false

=item retain_underscore => bool

If false, underscores are removed from search words.
Default: false

=item others

All other options are passed to components of a WordNet object.
(See: L<WordNet::IO::Disk>, L<WordNet::Filenames>, etc.)

=head1 SEE ALSO

L<WordNet::IO::Disk>, L<WordNet::IO::Cache>, L<WordNet::IO::XML>

L<WordNet::Filenames>

=head1 AUTHOR

Ben Haskell, E<lt>ben@clarity.princeton.eduE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 Princeton University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
