package WordNet::Filenames;
use 5.006;
use strict;
use warnings;
use WordNet qw/:pos/;

require Exporter;

our @ISA = qw/Exporter/;
our (%EXPORT_TAGS, @EXPORT_OK, @EXPORT) = ();
our $VERSION = '0.01';

use File::Spec;

my $PATH_SEP = qr/(?:(?<!\b\w):(?!\\))|(?:;(?=\w:|\.?[\\\/]))/;

our @defaults = ( userdir => $ENV{HOME} );

sub new {
    my $class = shift;
    my $ref = { @defaults, @_ };
	my @order = (
		[ wnsearchdir => $ref ],
		[ wnsearchpath => $ref ],
		[ wnpath => $ref ],
		[ wnhome => $ref ],
		[ WNSEARCHDIR => \%ENV ],
		[ WNSEARCHPATH => \%ENV ],
		[ WNPATH => \%ENV ],
		[ WNHOME => \%ENV ],
	);
	my @search;
	for (@order) {
		my ($key, $hash) = @$_;
		my @dirs = ($hash ? $hash->{$key} : $key);
		next unless grep $_, @dirs;
		@dirs = map split($PATH_SEP), @dirs if $key =~ /path/i;
		unshift @dirs, map File::Spec->catfile($_, 'dict'), @dirs;
		push @search, $_ for grep -d, @dirs;
	}
	my %seen;
    $ref->{search} = [ grep !$seen{$_}++, @search ];
	$ref->{wnsearchdir} ||= $ref->{search}[0];
    my $self = bless $ref, $class;
	$self->_resolve_dirs;
	$self;
}

sub _resolve_dirs {
	my $self = shift;
	my %done = map { $_ => 0 } grep $self->{$_}, grep /dir$/, keys %$self;
	my $no_new = 0;
	while ($no_new < 2 and grep !$_, values %done) {
		my $c = grep $_, values %done;
		for (keys %done) {
			next if $done{$_};
			next unless -d $self->{$_};
			$done{$_} = 1;
		}
		for (keys %done) {
			next unless exists $done{$self->{$_}};
			next unless $done{$self->{$_}};
			$self->{$_} = $self->{$self->{$_}};
		}
		$no_new++ if $c == grep $_, values %done;
	}
}

sub catfile {
    my $self = shift;
    my ($dir, $filename) = (@_);
	$dir .= 'dir' unless $dir =~ /dir$/;
    File::Spec->catfile($self->{$dir}, $filename);
}

sub find {
    my $self = shift;
    for my $d (@{$self->{search}}) {
        for (@_) {
            my $fn = File::Spec->catfile($d, $_);
			next unless -f $fn;
			$self->{_lastfound_dir} = $d;
			$self->{_lastfound_file} = $_;
            return $fn;
        }
    }
    return '';
}
sub _find_in_same_dir {
	my ($s, $key, @f) = @_;
	my $last_time = $s->{_lastfound_dirs}{$key} || '';
	return '' unless my $file = $s->find(@f);
	my $this_time = $s->{_lastfound_dir} || '';
	$s->{_lastfound_dirs}{$key} = $this_time;
	if ($last_time and $last_time ne $this_time) {
		$s->{_lastfound_dirs}{$key} = $this_time if $this_time;
		return '';
	}
	return $file;
}
sub data {
	my ($s, $w) = @_;
	$w = $pos_medium{$w} if exists $pos_medium{$w};
	$s->_find_in_same_dir('data/index', "data.$w", "$w.dat");
}
sub index {
	my ($s, $w) = @_;
	$w = $pos_medium{$w} if exists $pos_medium{$w};
	$s->_find_in_same_dir('data/index', "index.$w", "$w.idx");
}

sub dir {
    my $self = shift;
    my ($filename) = (@_);
    File::Spec->catdir(File::Spec->splitpath($filename));
}

sub root {
    my $self = shift;
    $self->catfile('root', @_);
}

sub home {
    my $self = shift;
    $self->catfile('wnhome', @_);
}
sub wnhome { home(@_); }

sub dict {
    my $self = shift;
    $self->catfile('wnsearch', @_);
}
sub search   { dict(@_); }
sub wnsearch { dict(@_); }

sub user {
    my $self = shift;
    $self->catfile('user', @_);
}

1;
__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

WordNet::Filenames - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::Filenames;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for WordNet::Filenames, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.


=head1 AUTHOR

A. U. Thor, E<lt>a.u.thor@a.galaxy.far.far.awayE<gt>

=head1 SEE ALSO

L<perl>.

=cut
