package WordNet::IO::XML;
use strict;
use warnings;

use base 'WordNet::IO';
our (%EXPORT_TAGS, @EXPORT_OK, @EXPORT);
%EXPORT_TAGS = @EXPORT_OK = @EXPORT = ();

our $VERSION = '0.01';

use 5.006;

use Carp;

sub escape { local $_ = shift; s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g; $_ }

sub start_tag {
    my ($name, @attr) = @_;
    my $tag = "<$name";
    while (my ($a,$v) = splice @attr, 0, 2) {
        $tag .= " $a=" . join '"', "", escape($v), "";
    }
    $tag .= ">";
    $tag;
}

sub empty_tag { my $tag = start_tag(@_); $tag =~ s!>$! />!; $tag }
sub end_tag   { my $tag = shift;         $tag = "</$tag>";  $tag }

sub data_element {
    my ($name, $data, @attr) = @_;
    my $tag = start_tag($name, @attr);
    $tag .= $data; # cdata($data);
    $tag .= end_tag($name);
    $tag;
}

sub cdata {  "<![CDATA[$_[0]]]>"; }
    
sub write {
    my $self = shift;
    my ($synset) = @_;
    my $text = start_tag("synset",
                         id => $synset->id,
                         pos => $synset->pos,
                         sk => $synset->sk,
                         (defined $synset->{frames} and scalar @{$synset->{frames}}
                          ? (frames => join(",", @{$synset->{frames}}))
                          : ())
                         );
    foreach my $word (@{$synset->{words}}) {
        $text .= start_tag("word",
                           word => $word->word,
                           id => $word->id,
                           pos => $word->pos,
                           sk => $word->sk,
                           ($word->syntax
                            ? (syntax => $word->{syntax})
                            : ()),
                           ($word->lexsense
                            ? (lex_id => $word->{lexsense})
                            : ()));
        foreach my $ptr ($word->pointers) {
            $text .= empty_tag("pointer",
                               to => $ptr->{to}{id},
                               pos => $ptr->{to}{pos},
                               type => $ptr->type);
        }
        $text .= empty_tag("frame", num => $_) for $word->frames;
        $text .= end_tag("word");
    }
    foreach my $ptr ($synset->pointers) {
        $text .= empty_tag("pointer",
                           to => $ptr->{to}{id},
                           pos => $ptr->{to}{pos},
                           type => $ptr->type);
    }
    $text .= empty_tag("frame", num => $_) for $synset->frames;
    if ($synset->gloss) {
        foreach my $type (keys %{$synset->{gloss}}) {
            $text .= data_element("gloss",
                                  ( ($type eq 'original')
                                    ? cdata($synset->gloss($type))
                                    : $synset->gloss($type) ),
                                  type => $type);
        }
    }
    $text .= end_tag("synset");
    $text;
}

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

WordNet::IO::XML - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::IO::XML;
  blah blah blah

=head1 ABSTRACT

  This should be the abstract for WordNet::IO::XML.
  The abstract is used when making PPD (Perl Package Description) files.
  If you don't want an ABSTRACT you should also edit Makefile.PL to
  remove the ABSTRACT_FROM option.

=head1 DESCRIPTION

Stub documentation for WordNet::IO::XML, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Ben Haskell, E<lt>bhaskell@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Ben Haskell

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
