package WordNet::Morphology::Simple;
## THIS IS NOT MAINTAINED!!!!
use strict;

our @defaults = ( filename => 'simple.cached',
                  location => '/wordnet/wn/dict',
                  inf_loc  => '/wordnet/wn/autotag/various',
                  ifile => 'inflect.tbl',
                  dfile => 'deinflect.tbl' );

my $filename;
my $WN_LOC;
my ($ifile, $dfile);

my %fname = ();


our $VERSION = '0.01';

my @pos;
BEGIN { @pos = qw/noun verb adj adv/; }

sub import {
    my @EXPORT = qw/moan groan isword/;
    my $package = shift;
    my ($caller, $file, $line) = caller();

    use File::Spec;

    my %param = (@defaults, @_);
    ($filename, $WN_LOC, $ifile, $dfile) =
        @param{"filename","location","ifile","dfile"};
    foreach my $p (@pos) {
        $fname{$p}{irr} = File::Spec->catfile($param{inf_loc}, "$p.irr");
    }
    $ifile = File::Spec->catfile($param{inf_loc}, $ifile);
    $dfile = File::Spec->catfile($param{inf_loc}, $dfile);

    no strict 'refs';
    foreach my $export (@EXPORT) {
        *{"$caller\::$export"} = *{"$package\::$export"};
    }
    &InitEverything;
    load_cached() or create_cached();
    1;
}

my $no_checks = 0;
my $case_insensitive = 1;

my @prepositions = qw(off on in out under over up down to from
                      of at with into for about between);
my @determiners = qw(a an the);

my %pos_to_char = qw/noun n verb v adj a adv r/;
my %char_to_pos = reverse %pos_to_char;
my %pos_base = qw/noun ns verb vp adj ab adv rb/;
my %substitutions;
my @special_subs = qw/() $1 \1 $1/;

my %inflections;
my %deinflections;

my @default_search = @pos;

my %isa_word = ();

my %sub_parts = (n => [ qw/ns np ng npg/ ],
                 v => [ qw/vp vs vps vpp ving/ ],
                 a => [ qw/ab ac as/ ],
                 r => [ qw/rb rc rs/ ]);
foreach my $k (keys %sub_parts) {
    foreach my $j (@{$sub_parts{$k}}) {
        $sub_parts{$j} = [ $j ];
    }
}
foreach my $p (@pos) {
    $sub_parts{$p} = $sub_parts{$pos_to_char{$p}};
}

my %part_order;
{ my @parts = ();
  foreach my $k (qw/n v a r/) {
      push @parts, @{$sub_parts{$k}};
  }
  my $j = 1;
  foreach my $part (@parts) {
      $part_order{$part} = $part_order{"$part|1"} = $j++;
      $part_order{"$part|0"} = $j++;
  }
}

my %super_part = qw/n noun v verb a adj r adv/;
foreach my $pos (keys %super_part) {
    foreach my $sp (@{$sub_parts{$pos}}) {
        $super_part{$sp} = $super_part{$pos};
    }
}
foreach my $pos (@pos) {
    $super_part{$pos} = $pos;
}

sub LoadINFL {
    my %so_far = ();
    my %seen_already = ();
    my $index;
    my ($pos, $pat, $rep, $to_pos);
    my @t_ref = (\%inflections, \%deinflections);
    my @f_name = ($ifile, $dfile);
    foreach my $table (@t_ref) {
        my $f_name = shift(@f_name);
        open FILE, $f_name
            or open FILE, "$WN_LOC/$f_name"
            or die "LoadINFL $f_name";
        %so_far = ();
        %seen_already = ();
        local $_;
        while (<FILE>) {
            next if (/^\#/ or /^\s*$/);
            if (/^s\s+([A-Z])\s+(\S+)\s*$/) {
                $substitutions{$1} = $2;
                next;
            }
            if (($pos, $pat, $rep, $to_pos) =
                /^(\w+)\s+(\S+)\s+(\S+)\s+(\w+)\s*$/) {
                $pat .= '$';
                if (exists($seen_already{$pos}->{$pat})) {
                    $index = $seen_already{$pos}->{$pat};
                } else {
                    $index = $so_far{$pos}++;
                    $seen_already{$pos}->{$pat} = $index;
                    $table->{$pos}->[$index]->{'pattern'} = $pat;
                }
                push(@{$table->{$pos}->[$index]->{'replacements'}},
                     { pattern => $rep,
                       to_pos  => $to_pos });
                next;
            }
            print STDERR "Malformed line in $f_name:\n$_";
        }
        close(FILE);
    }
    foreach my $table (\%inflections, \%deinflections) {
        foreach my $p (keys %$table) {
            for (my $c = 0; $c < @{$table->{$p}}; $c++) {
                my $subbed = $table->{$p}->[$c]->{'pattern'};
                foreach my $s (keys %substitutions) {
                    $subbed =~ s/$s/$substitutions{$s}/g;
                }
                $table->{$p}->[$c]->{'subbed'} = $subbed;
                for (my $d = 0; $d < @{$table->{$p}->[$c]->{'replacements'}}; $d++) {
                    my $subbed = $table->{$p}->[$c]->{'replacements'}->[$d]->{'pattern'};
                    foreach my $s (keys %substitutions) {
                        $subbed =~ s/$s/$substitutions{$s}/g;
                    }
                    my @subs = @special_subs;
                    while (@subs) {
                        my $s = shift(@subs);
                        my $r = shift(@subs);
                        $subbed =~ s/\Q$s\E/$r/;
                    }
                    $table->{$p}->[$c]->{'replacements'}->[$d]->{'subbed'} = $subbed;
                }
            }
        }
    }
}

my %exc_infl;
my %exc_base;

sub LoadIRR {
    %exc_infl = ();
    %exc_base = ();
    my %formats = (verb => { 3 => [ qw(vp vps vpp) ],
                             7 => [ qw(vp vs vs vp vps vps vpp) ] },
                   noun => { 2 => [ qw(ns np) ] },
                   adj => { 2 => [ qw(ignoreab ignoreai) ] ,
                            3 => [ qw(ab ac as) ] },
                   adv => { 2 => [ qw(ignorerb ignoreri) ],
                            3 => [ qw(rb rc rs) ] });
    foreach my $pos (@pos) {
        my $fname = $fname{$pos}{irr};
        open FILE, $fname
            or die "LoadIRR $fname";
        local $_;
        while (<FILE>) {
            next if (/^\#/ or /^\s*$/);
            chomp;
            my ($line, $prefixes) = split(/\s:\s/);
            my @split = split(' ',$line);
            foreach my $str (@split) {
                $str =~ s/_/\ /g;
            }
            my @prefixes = ("");
            if (defined($prefixes)) {
                push(@prefixes, split(' ',$prefixes));
            }
            foreach my $pre (@prefixes) {
                my @actual = map { $pre . $_; } @split;
                if (!defined($formats{$pos}->{scalar(@split)})) {
                    print STDERR "Invalid IRR file line: ",
                    join(" ", @split), "\n";
                    last;
                }
                my @forms = @{$formats{$pos}->{scalar(@split)}};
                my $base = shift(@actual);
                my $base_form = shift(@forms);
                while (@actual) {
                    my $act = shift @actual;
                    my $frm = shift @forms;
                    push @{$exc_infl{$frm}->{$act}}, $base, $base_form;
                    push @{$exc_base{$base_form}->{$base}}, $act, $frm;
                }
            }
        }
        close(FILE);
    }
}

sub InitEverything {
    LoadINFL;
    LoadIRR;
}

sub load_cached {
    open FILE, "<$filename" or return 0;
    # die "Couldn't open  cache file: $filename. $@";
    local $_;
    while (<FILE>) {
        chomp;
        my($pos,$w) = split //, $_, 2;
        $w = lc $w if $case_insensitive;
        $isa_word{$pos}->{$w} = 1;
    }
    close FILE;
    return 1;
}

sub create_cached {
    print STDERR "Creating cache.\nSubsequent uses will not incur this delay.\n...";
    use File::Spec;
    my $longfn;
    if ($^O =~ /(MSWin|OS2|DOS)/i) {
        $longfn = 0;
    } else {
        $longfn = 1;
    }
    open FILE, ">$filename" or die "Couldn't create cache file: $filename. $@";
    my %words = ();
    foreach my $pos (@pos) {
        my $fname = File::Spec->catfile($WN_LOC, (($longfn) ? "data.$pos" : "$pos.dat"));
        open my $fh, $fname or die "Couldn't open: $fname. $@";
        local $_;
        while (<$fh>) {
            next unless /^\d/;
            my @dat = split /\s/;
            my ($pos, $count) = ($dat[2], hex $dat[3]);
            splice @dat, 0, 4;
            my @words = splice @dat, 0, $count*2;
            while (@words) {
                my $word = shift @words;
                shift @words;
                $word =~ s/\(..?\)$//;
                $word =~ s/_/\ /g;
                $words{"$pos$word"} = 1;
            }
        }
    }
    foreach my $key (%words) { print FILE "$key\n"; }
    close FILE;
    print STDERR "Done\n";
    load_cached();
}

sub misa { $a = shift; return $pos_to_char{$super_part{$a}}; }
sub mexc { $a = shift; return $a; }
sub moan  {
    my @ret = analyze(\%deinflections, \%exc_infl, @_);
    my @actual = ();
    while (@ret) {
        my ($w, $p) = splice @ret, 0, 2;
        if ($no_checks) {
            $p  .= IS_A_WORD($w, $p, \%exc_infl, \&misa, \&mexc)
                ? "|1"
                : "|0";
            push @actual, $w, $p;
        } else {
            push @actual, $w, $p
                if IS_A_WORD($w, $p, \%exc_infl, \&misa, \&mexc);
        }
    }
    return @actual;
}

sub gisa { $a = shift; return $pos_to_char{$super_part{$a}}; }
sub gexc { $a = shift; return $pos_base{$super_part{$a}}; }
sub groan {
    my ($word, @parts) = @_;
    @parts = @default_search if not @parts;
    @parts = map { IS_A_WORD($word, $_, \%exc_base, \&misa, \&mexc)
                       ? ($_)
                       : (); } @parts unless $no_checks;
    my @ret = @parts ? analyze(\%inflections, \%exc_base, $word, @parts) : ();
    return @ret unless $no_checks;
    my @actual = ();
    while (@ret) {
        my ($w, $p) = splice @ret, 0, 2;
        $p .= IS_A_WORD($word, $p, \%exc_base, \&gisa, \&gexc)
            ? "|1"
            : "|0";
        push @actual, $w, $p;
    }
    return @actual;
}

sub IS_A_WORD {
    my ($w, $p, $exc, $ismap, $excmap) = @_;
    my ($ipos, $epos) = map {$_->($p)} ($ismap, $excmap);
#    $w =~ s/\x00//g;
    return 1 if $isa_word{$ipos}->{$w};
#    return 1 if $exc->{$epos}->{$w};
    return 0 if not $case_insensitive;
    return 0 if $w eq lc $w;
    return IS_A_WORD(lc $w, $p, $exc, $ismap, $excmap);
}

sub isword {
    IS_A_WORD($_[0], $pos_to_char{$super_part{$_[1]}}, \%exc_base, \&misa, \&mexc);
}

###### START COPY PASTED

sub preserve_case {
    my ($old, $new) = @_;
    my $mask = uc $old ^ $old;
    $mask .= substr($mask, -1) x (length($new) - length($old));
    $mask = substr($mask,0,length($new));
    uc $new | $mask;
}

sub analyze {
    my ($table, $exc, $word, @to_search) = @_;
    @to_search = @default_search if not @to_search;
    @to_search = map { ((defined $sub_parts{$_})
                        ? (@{$sub_parts{$_}})
                        : ()); } (map { $pos_to_char{$_}; } @to_search);

    my @ret = ();

    $word =~ s/_/\ /g;

    return @ret if not $word;

    my $lcword = lc $word;
    my @parts = split /([- ])/, $word;
    my @lcparts = map lc, @parts;
    my $first = first_prep_or_det(@parts);
    my @to_check = ();

    my %found_exc = ();
    foreach my $spos (@to_search) {
        my @to_add = ();
        push @to_add, @{$exc->{$spos}->{$lcword}} if defined $exc->{$spos}->{$lcword};
        push @to_add, @{$exc->{$spos}->{$word}} if defined $exc->{$spos}->{$word};
        while (@to_add) {
            my ($w, $p) = splice @to_add, 0, 2;
            push @ret, preserve_case($word, $w), $p;
            $found_exc{$p}++;
        }
        if ($spos eq 'ns' or $spos eq 'np') {
            if (@parts == 1) {
                @to_check = ([0]);
            } elsif (@parts == 5 and $parts[2] =~ /^and$/i) {
                @to_check = ([0,4], [4]);
            } elsif ($first != @parts) {
                if ($first > 0 and $first != $#parts) {
                    @to_check = ([$first - 2]);
                } elsif ($first > 0) {
                    if (@parts == 3) {
                        @to_check = ([0], [2]);
                    } else {
                        @to_check = ([$#parts]);
                    }
                } else {
                    @to_check = ([$#parts]);
                }
            } else {
                if (@parts == 3) {
                    @to_check = ([0],[2]);
                } else {
                    @to_check = ([$#parts]);
                }
            }
        } elsif ($super_part{$spos} eq 'verb') {
            if (@parts == 1) {
                @to_check = ([0]);
            } elsif (@parts == 5 and $parts[2] =~ /^and$/i) {
                @to_check = ([0,4]);
            } elsif ($first != @parts) {
                if (!$first) {
                    @to_check = ([2]);
                } else {
                    @to_check = ([0]);
                }
            } else {
                @to_check = ([0]);
            }
        } else {
            if (@parts > 1) {
                @to_check = ([0], [$#parts]);
            } else {
                @to_check = ([0]);
            }
        }
        foreach my $aref (@to_check) {
            my %forms = ();
            foreach my $i (@$aref) {
                my @analyzed = analyze_word($table->{$spos},
                                            $exc->{$spos},
                                            $parts[$i]);
                while (@analyzed) {
                    my $form = shift @analyzed;
                    push @{$forms{shift @analyzed}->{$i}}, $form;
                }
            }
            foreach my $new_pos (keys %forms) {
                my @to_combine = map { [ $_ ]; } @parts;
                next if scalar(keys %{$forms{$new_pos}}) != @$aref;
                foreach my $i (@$aref) {
                    $to_combine[$i] = [ @{$forms{$new_pos}->{$i}} ];
                }
                my @results = combine(@to_combine);
                foreach my $result (@results) {
                    push @ret, $result, $new_pos if not $found_exc{$new_pos};
                }
            }
        }
    }
    my @actual_results = ();
    my %unique = ();
    while (@ret) {
        my ($w, $p) = splice @ret, 0, 2;
        $unique{$part_order{$p}}->{$p}->{$w} = 1;
    }
    foreach my $ord (sort keys %unique) {
        foreach my $p (keys %{$unique{$ord}}) {
            foreach my $w (sort keys %{$unique{$ord}->{$p}}) {
                push @actual_results, $w, $p;
            }
        }
    }
    return @actual_results;
}

sub combine {
    my @arrays = @_;
    if (!@arrays) { return (); }
    my @array = @{shift(@arrays)};
    my @ret = ();
    foreach my $word (@array) {
        next if (!defined($word));
        my @suffixes = combine(@arrays);
        if (@suffixes) {
            foreach my $suffix (@suffixes) {
                next if (!defined($suffix));
                push(@ret, "$word$suffix");
            }
        } else {
            push(@ret, "$word");
        }
    }
    return @ret;
}

sub analyze_word {
    my ($table, $exc, $word) = @_;
    my @ret = ();
    my %found_exc = ();
    if (defined($exc->{$word})) {
        my @exc = @{$exc->{$word}};
        while (@exc) {
            my ($w, $p) = splice(@exc, 0, 2);
            push(@ret, $w, $p);
            $found_exc{$p} = 1;
        }
    }
    foreach my $pref (@$table) {
        my $pat = $pref->{'subbed'};
        foreach my $rref (@{$pref->{'replacements'}}) {
            my $rep = $rref->{'subbed'};
            my $new_word = apply_infl($word, $pat, $rep);
            if (defined($new_word)) {
                my $new_pos = $rref->{'to_pos'};
                push(@ret, $new_word, $new_pos) if (!$found_exc{$new_pos});
            }
        }
    }
    return @ret;
}

sub apply_infl {
    my ($word, $pattern, $replacement) = @_;
    my $new_word = undef;
    if ($word =~ $pattern) {
        $new_word = $word;
        my @subs = ($1, $2, $3, $4);
        for (my $i = 1; $i <= @subs; $i++) {
            $replacement =~ s/\$$i/$subs[$i-1]/g;
        }
        $new_word =~ s/$pattern/$replacement/;
    }
    return $new_word;
}

sub first_prep_or_det {
    my @word = @_;
    for (my $i=0; $i<@word; $i++) {
        foreach my $pd (@prepositions, @determiners) {
            return $i if ($pd eq $word[$i]);
        }
    }
    return scalar(@word);
}

###### END COPY PASTED

1;
__END__
# TODO - Edit Documentation

=head1 NAME

WordNet::Morphology::Simple - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::Morphology::Simple;
  blah blah blah

=head1 ABSTRACT

  This should be the abstract for WordNet::Morphology::Simple.
  The abstract is used when making PPD (Perl Package Description) files.
  If you don't want an ABSTRACT you should also edit Makefile.PL to
  remove the ABSTRACT_FROM option.

=head1 DESCRIPTION

Stub documentation for WordNet::Morphology::Simple, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

moan, groan, isword


=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Ben Haskell, E<lt>bhaskell@clarity.princeton.eduE<gt>

=head1 COPYRIGHT AND LICENSE

TODO -- Insert WordNet License

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
