#!/usr/bin/perl -w
use strict;

use FileHandle;

my ($pos, $pat, $rep, $new, $group, $suffix, $n, $m, $r);
my $infile = new FileHandle;
my $outforward = new FileHandle;
my $outbackward = new FileHandle;
my $subs = new FileHandle;
open($infile, "inflections.tbl") or die "Couldn't open inflections table.\n";
open($outforward, ">inflect.tbl") or die "Couldn't open inflection table for writing.\n";
open($outbackward, ">deinflect.tbl") or die "Couldn't open de-inflection table for writing.\n";
open($subs, ">subs.tbl") or die "Couldn't open subs table for writing.\n";

my $notsilent = 1;
foreach (@ARGV) { if (/silent/) { $notsilent = 0; } }

my $line_no = 0;
while (<$infile>) {
    $line_no++;
    if (/^\#/) {
	print $outforward $_;
	print $outbackward $_;
    }
    elsif (/^s\s*[A-Z]\s*\S*/) {
	print $outforward $_;
	print $outbackward $_;
	print $subs $_;
    }
    elsif (/^(\w+)\s+\{?([^\}\s]+)\}?\s+\{?([^\}\s]+)\}?\s+(\w+)\s*$/) {
	($pos, $pat, $rep, $new) = ($1, $2, $3, $4);
	$pat =~ s/\%//g;
	$rep =~ s/\$1/\(\)/;
	if ($pat =~ /^[^\(\)]*\((.*)\)[^\(\)]*$/) {
	    $group = $1;
	    $rep =~ s/\(\)/\($group\)/;
	}
	my $n = $pat;
	$n =~ s/^([^\(\)]*)\(.*\)([^\(\)]*)$/$1\(\)$2/;
	my $m = $rep;
	$m =~ s/^([^\(\)]*)\(.*\)([^\(\)]*)$/$1\(\)$2/;
	print $outforward "$new\t$rep\t$n\t$pos\n";
	print $outbackward "$pos\t$pat\t$m\t$new\n";
	print "$new:$rep -> $pos:$pat | $pos:$pat -> $new:$rep\n" if ($notsilent);
    }
    elsif ($_ !~ /\s*/) {
	print "Error in line $line_no: $_";
    }
}
close($infile);
close($outforward);
close($outbackward);
close($subs);
