#!/usr/local/bin/perl -T

#
# modified 17 June 2005 for mod_perl - chandra@clarity
# 
# STUFF TO DO
# Some names in the old wnb aren't the same as in the WordNet module
#     (e.g. stem adjective vs. pertainym).  Also, some names could be
#     clearer (such as 'domain').
# Should the user be able to display descriptions next to relations
#     (e.g. "hypernyms (is a kind of ...)")?

#BEGIN { $ENV{WNSEARCHDIR} = '/wordnet/wn/src/Release/3.0/dict'; }
BEGIN { $ENV{WNSEARCHDIR} = '/var/local/as/home/cslweb/WordNet/3.0/dict/'; }
our ($title, $version, $buglink, $homeLink, $disable_advanced);
BEGIN {
    $version = '3.0';
    $buglink = 'http://wordnet.princeton.edu/cgi-bin/bugsubmit.pl';

    $homeLink = "<a href=\"http://wordnet.princeton.edu\">WordNet home page</a>";
    $title = "WordNet Search - $version";
    $disable_advanced = 1;
}
use strict;
use warnings;
#use lib '/wordnet/wn/2k3/lib/perl';
use lib '.';     # for node.pm
#use lib '/usr/csl/apache/perl/mptest';     # for node.pm
use lib './mptest';     # for node.pm
use WordNet qw/:vars/;
use node;
use CGI qw/:standard *ol *ul *div/;
our $self_url = url(-relative => 1);
$CGI::POST_MAX = 1024;  # max 1K posts
$CGI::DISABLE_UPLOADS = 1;  # no uploads

my $r = shift;	# get Apache::RequestRec object

# main functions:
# all this will be wrapped in a handler, so we need to make "my" references sharable with "our" when they are used in a named sub.
# sub checkParams; # launders tainted parameters
# sub buildHistory;
# sub buildR;      # used by buildHistory (recursive)
# sub printForm;
# sub printHistory;
# sub printR;      # used by printHistory (recursive)
# sub longHist;    # used by printR (does actual HTML & formatting)

# utility functions:
# sub colorpos;   # takes Word/Synset object, returns colored HTML for pos
# sub makeLink;   # takes new $histInfo & (optional) setting name, returns link
# sub escHTML;    # takes link and changes spaces to '+'
# sub splitGloss; # takes gloss, returns array: (domain, def, sentences)
# sub printError; # takes error message, prints it, & exits
# sub fitBrowser; # returns string that should be used for CSS acc. to browser

# Java Script
our $JSCRIPT=<<END;
function sf(){document.f.s.focus();}
END

#DEBUGGING: set values to 1 to print out debugging info, 0 to not
our %DEBUG = qw/general 0 cathist 0 buildHistory 0 printHistory 0 longHist 0
    makeLink 0/;

our $MAX_HISTORY_LENGTH = 1000; # max number of lines in results
our $MAX_SEARCH_LENGTH = 500;  # max length for search term

our %querydataR = reverse(%querydata);
#querydata maps "@" to "hype"
#querydataR maps "hype" to "@"
#english maps "@" to "hypernym"
our $pos = 'nvars';
our %shortPos = qw/n n v v a adj r adv s sat/;

#Extensions to WordNet database relations:
our %extRels = %node::extRels;

#hashes of relations which allow recursive (show all) searches
our %recRels1 = qw/hype 1 hmem 1 hprt 1 hsub 1 holo 1/;
our %recRels2 = qw/hypo 1 mmem 1 mprt 1 msub 1 mero 1/;

# Script Parameters:
#
# explicitly shared for wrapping in handler
# take them from the request record object
#


#our $h_url = ($r->method eq 'GET') ? 1 : 0; #tests whether $h came from form or query string

our $searchTerm = param('s');   #word being searched for (root word)
our $h = param('h');            #history list (list of digits [0-3]*)
$h = ($h =~ /^([0-3]*)$/) ? $1 : '';
our $histInfo = param('i');     #index in history list
$histInfo = ($histInfo =~ /^(-?\d+)$/) ? $1 : '';
our $recSearch = param('r');    #(1=showAll, 2=extended [e.g. coordinate])
our $jump = param('j');         #index in history list to jump to
our $change_opt = param('c');   #which option to change
our $submit = param('sub');     #tells which submit button was clicked
$submit = ($submit eq 'Search WordNet') ? 'Search WordNet' : ($submit eq 'Change') ? 'Change' : undef;

our $count = 0;                 # number of lines in result
                                # defined out here due to weird mod_perl scoping

# DISPLAY OPTIONS:
our @options = ( example  => 1, 'Example Sentences',
                gloss    => 1,  'Glosses',
                count    => '', 'Frequency Counts',
                offset   => '', 'Database Locations',
                lex      => '', 'Lexical File Info',
                lexnum   => '', 'Lexical File Numbers',
                sensekey => '', 'Sense Keys',
                sensenum => '', 'Sense Numbers',
                );
our (%options, %opt_desc, %option_num);
{
    our $fields = 3;
    for our $i (grep { not $_ % $fields } 0..$#options) {
        $options{$options[$i]} = $options[$i+1];
        $opt_desc{$options[$i]} = $options[$i+2];
        $option_num{$options[$i]} = $i / $fields;
    }
}
our %default_opt = %options;
our %num_option = reverse %option_num;

for our $k (keys %options) {
    our $p = param('o'.$option_num{$k});    
    $options{$k} = $p if defined $p;
}

our $history = '';   # holds base (empty) node of history tree
our @h;         # history list (0 = not expanded, 1 = expanded,
               #   2 = suppressed/not expanded, 3 = suppressed/expanded)
our $warning;   # if ne '', prints a warning at the top of the search page

print header;
print start_html(-title=>$title,
                 -style=>{-code=>fitBrowser()},
                 -script=>$JSCRIPT,
                 -onLoad=>((defined($histInfo) || defined($jump))
                           ? '' : 'sf()'));
# printf("\$searchTerm = %s\n\$h = %s\n\$histInfo = %s\n\$recSearch = %s\n\$jump = %s\n\$change_opt = %s\n\$submit = %s\n\$h_url = %s\n", $searchTerm, $h, $histInfo, $recSearch, $jump, $change_opt, $submit, $h_url);

checkParams();

$h = '' if $submit eq 'Search WordNet'; # or not $h_url;

if ($change_opt != -1 and $num_option{$change_opt}) {
    our $change = $num_option{$change_opt};
    $options{$change} = $options{$change} ? '' : 1;
}
our $LINKS_OPTS = ''; # this string is used many times by makeLink
for our $k (keys %options) {
    $LINKS_OPTS .= 'o' . $option_num{$k} . '=' . $options{$k} . '&';
}
our $jumptoo = $jump - 4;  # jump is bold; jumptoo is location on page
if ($jumptoo < 0) { $jumptoo = 0; }

@h = split //, $h;

buildHistory();

@h = split //, $h;

$DEBUG{'general'} and print p("after buildHistory, h has "
                              . length($h) . " items: $h");
$DEBUG{'cathist'} and $history->printdebug;

print start_div({class=>'header'});
print span({class=>'title'}, $title);
print ' - ', $homeLink, ' - ';

    print a({href=>'advanced.pl'}, "Advanced Search"),
    ' - ' unless $disable_advanced;
    print a({href=>'http://wordnet.princeton.edu/man/wngloss.7WN'},
      "Glossary"),
    ' - ',
#    a({href=>'http://wordnet.princeton.edu/webwn.help'}, "Help") . "\n";
    a({href=>'mptest/help.pl'}, "Help") . "\n";
print end_div;
printForm();

printHistory();

print $homeLink;
print end_html;


sub checkParams {
    if (defined($change_opt)) {
        if ($change_opt =~ m/^([\-]?[0-9]*)$/) {
            $change_opt = $1;
            if ($change_opt eq '') { $change_opt = -1 }
        } else {
            printError("The \'o\' parameter is corrupted."
                       . " Please restart your search.");
        }
    } else { $change_opt = -1 }
    for my $k (keys %options) {
        if (defined($options{$k})) {
            if ($options{$k} =~ m/^(1?)$/) {
                $options{$k} = $1;
            } else {
                printError("The \'o$option_num{$k}\' parameter is corrupted."
                           . " Please restart your search.");
            }
        } else { $options{$k} = ''; }
    }
    if (defined($recSearch)) {
        if ($recSearch =~ m/^([12]?)$/) {
            $recSearch = $1;
        } else {
            printError("The \'r\' parameter appears to have been corrupted."
                        . " (recSearch=" . $recSearch . ")"
                       . " Please restart your search.");
        }
    } else { $recSearch = 0; }
    if (defined($h)) {
        if (($h =~ m/^([0-3]*)$/) and length($h) <= $MAX_HISTORY_LENGTH) {
            $h = $1;
        } else {
            printError("The \'h\' parameter appears to have been corrupted."
                        . " (h=" . $h . ")"
                       . " Please restart your search.");
        }
    } else { $h = ''; }
    if (defined($searchTerm)) {
        if ($searchTerm =~ m/^([a-zA-Z0-9\-\. \/\']*)$/) {
            $searchTerm = $1;
        } else {
            printError("Sorry, your search can only contain "
                       . "letters, numbers, spaces, hyphens, periods, "
                       . "slashes, and/or apostrophes.");
        }
    } else { 
        $searchTerm = '';
    }
    if ($searchTerm eq '') {
#        if (defined($histInfo)) {
#            printError("The \'s\' and \'i\' parameters appear to have been"
#                       . " corrupted. Please restart your search.");
#        }
        $histInfo = -1;
    }
    $searchTerm =~ s/^ *//g;
    $searchTerm =~ s/ *$//g;
    $searchTerm =~ s/ +/ /g;
    if (defined($histInfo)) {
        if ($histInfo =~ m/^(\-?[0-9]*)$/ && $histInfo >= -1) {
            $histInfo = $1;
            if ($histInfo eq '') { $histInfo = -1; }
        } else {
            printError("The \'i\' parameter appears to have been corrupted."
                        . " (histInfo=" . $histInfo . ")"
                       . " Please restart your search.");
        }
    } else { $histInfo = -1; }
    if ($jump) {
        if ($jump =~ m/^([0-9]*)$/) {
            $jump = $1;
        } else {
            printError("The \'j\' parameter appears to have been corrupted."
                        . " (jump=" . $jump . ")"
                       . " Please restart your search.");
        }
    } else { $jump = $histInfo; }
    if (defined($submit)) {
        if ($submit =~ m/^(Change|Search WordNet)$/) {
            $submit = $1;
        } else {
            printError("The \'sub\' parameter appears to have been corrupted."
                        . " (submit=" . $submit . ")"
                       . " Please restart your search.");
        }
    } else { $submit = ''; }
}


sub buildHistory {
    # build history
    return unless $searchTerm;
    $history = node->init($searchTerm);
    my @kids = @{$history->children};
    @kids or printError("Your search did not return any results.");
    @h = (0) x @kids unless @h;
    $count = 0; # iterator used by buildR to read along @h
    my $infoObject = undef;
    $count = buildR($count, $_, \$infoObject) + 1 for @kids;
    # expand/unsuppress history if needed
    if ($recSearch && !$recRels1{$infoObject->getobject}
        && !$recRels2{$infoObject->getobject}) {
        printError("The \'r\' parameter appears to have been corrupted."
                        . " (recSearch=" . $recSearch . ")"
                        . " recRels1 or recRels2 may be damaged."
                   . " Please restart your search.");
    }
    if ($histInfo > -1) {
        if ($histInfo > $#h) {
            printError("The \'i\' parameter appears to have been corrupted."
                       . " Please restart your search.");
        }
        if ($h[$histInfo] == 1 && !$recSearch && $h[$histInfo + 1] > 1) {
            $infoObject->unsuppress();
        } elsif ($infoObject->expand($recSearch, $MAX_HISTORY_LENGTH - @h + 1)
                < 0) {
                $warning = "The search exceeded the result limit, "
                    . "so the following list is valid but incomplete. "
                    . "Only the top levels of the list are displayed.";
            # else {
            #    if (!@{$infoObject->children}) {$warning = "no new results";}
            #}
        }
    }
    $history->finalize(-1);
    $h = $history->geth;
}


sub buildR {
    my ($count, $object, $infoObject) = @_;
    $$infoObject = $object if $count == $histInfo;
    if ($count >= $MAX_HISTORY_LENGTH || $count > $#h) {
        printError("The \'h\' parameter appears to have been corrupted."
                    . " $count >= $MAX_HISTORY_LENGTH || $count > $#h"
                   . " Please restart your search.");
    }
    if ($h[$count] == 1) {
        $object->expand(0);
        my @kids = @{$object->children};
        foreach (@kids) {
            $count++;
            $count = buildR($count, $_, $infoObject);
        }
    } elsif ($h[$count] == 2) {
        $object->suppressed(1);
    } elsif ($h[$count] == 3) {
        $object->suppressed(1);
        $object->expand(0);
        my @kids = @{$object->children};
        foreach (@kids) {
            $count++;
            $count = buildR($count, $_, $infoObject);
        }
    }
    return $count;
}


sub printForm {
    my %o_labels = ();
    $o_labels{-1} = '(Select option to change)';
    for my $k (keys %options) {
        $o_labels{$option_num{$k}} =
            ($options{$k} ? 'Hide ' : 'Show ').$opt_desc{$k};
    }

    print start_form(-method=>'GET', -action=>$self_url, -name=>'f'),
    "Word to search for: ",
    textfield({maxLength=>$MAX_SEARCH_LENGTH, name=>'s'}),
    submit(-name=>'sub', -value=>'Search WordNet');
    if ($warning) {
        print a({name=>'c'}, ''),
	        p(a({class=>'warning'}, 'WARNING: ' . $warning));
    }
    for my $k (keys %options) {
        param('o'. $option_num{$k}, $options{$k});
        print hidden('o'.$option_num{$k}, $options{$k});
    }

	# XSS?
	$h = ($h =~ /^([0-3]*)$/) ? $1 : '';
	# XSS?

    param('h', $h);
    print hidden('h', $h);
    param('s', $searchTerm);
    print end_form . "\n";

    print start_form(-method=>'GET', -action=>$self_url, -name=>'change') . "\n";
    param('c', -1);
    print "Display Options: ",
    popup_menu(-name=>'c',
               -values=>[-1, sort { $a <=> $b } values %option_num],
               -default=>-1,
               -labels=>\%o_labels), ' ',
    submit(-name=>'sub', -value=>'Change');
    for my $k (keys %options) {
        param('o'. $option_num{$k}, $options{$k});
        print hidden('o'.$option_num{$k}, $options{$k});
    }
    param('h', $h);
    print hidden('i', $histInfo), hidden('h', $h), hidden('s', $searchTerm);
    print end_form() . "\n";
}


sub printHistory {
    if (!$history) { return; }
    my @kids = @{$history->children};
    if (!@kids) { return; }
    my $curpos = $kids[0]->pos;
    print div({class=>'key'}, 'Key: "S:" = Show Synset (semantic) relations, "W:" = Show Word (lexical) relations') . "\n";
    print h3(ucfirst($pos_long{$curpos})) . "\n",
    start_ul;
    foreach (@kids) {
        if ($_->pos ne $curpos) {
            $curpos = $_->pos;
            print end_ul,
            h3(ucfirst($pos_long{$curpos})) . "\n",
            start_ul . "\n";
        }
        printR($_);
    }
    print end_ul . "\n";
}


sub printR {
    my $node = shift;
    my $count = $node->count;
    if ($count == $jumptoo && !$warning) {
        print a({name=>'c'}, '');  # jumps here when page loads
    }
    print li(longHist($node)) . "\n" unless $node->suppressed;
    my @kids = @{$node->children};
    return unless @kids;
    my $kidsSuppressed = $kids[0]->suppressed;
    my $curpos = undef;
    my $object = $node->getobject;
    unless ($kidsSuppressed) {
        if ($object eq 'over') {
            print start_ul . "\n";
            $curpos = $kids[0]->pos;
        } else {
            print start_ul({-class=>'type' . $kids[0]->type}) . "\n";
        }
    }
    foreach (@kids) {
        if ($object eq 'over' && $_->pos ne $curpos) {
            $curpos = $_->pos;
            print end_ul, start_ul;
        }
        printR($_);
    }
    unless ($kidsSuppressed) {
        if ($object eq 'over') { print end_ul . "\n"; }
        else { print end_ul . "\n"; }
    }
    return;
}


sub longHist {
    my $node = shift;
    my $object = $node->getobject;
    my $type = $node->type;
    my $i = $node->count;
    if ($DEBUG{'longHist'}) { print "longHist: i=$i, object=$object"; }
    my $histString = undef;
    if ($type eq 'r') { #relation
        if ($DEBUG{'longHist'}) { print " -> relation", p; }
        my @strings = ();
        if (exists($extRels{$object})) {
            $strings[0] = $extRels{$object};
        } elsif ($recRels1{$object}) {
            $strings[0] = 'direct ' . $english{$querydataR{$object}};
            $strings[1] = 'inherited ' . $english{$querydataR{$object}};
            if ($object eq 'hype') { $strings[2] = 'sister term' }
        } elsif ($recRels2{$object}) {
            if ($object eq 'hypo' && $node->parent->pos eq 'v') {
                $strings[0] = 'direct troponym';
                $strings[1] = 'full troponym';
            } else {
                $strings[0] = 'direct ' . $english{$querydataR{$object}};
                $strings[1] = 'full ' . $english{$querydataR{$object}};
            }
        } elsif ($object eq 'dummy') {
            my @prevWords = @{$node->relatedto};
            if (@prevWords) {
                $histString .= '[Applies to ' . join(', ',
                       map {a({href=>makeLink('', '', '', $_->word),
                               class=>'relWords'}, $_->word)}
                                                     @prevWords) . '] ';
            } else { $histString = '' }
            return $histString . $node->extra;
        } else {
            if ($object eq 'also' && $node->parent->pos eq 'v') {
                $strings[0] = 'phrasal verb';
            } else {
                $strings[0] = "<!-- $object ($type) --> ".$english{$querydataR{$object}};
            }
        }
        if ($i == $jump) {$strings[$recSearch] = b($strings[$recSearch])}
        $histString = i(a({href=>makeLink($i, $node)}, $strings[0]))
            . ($strings[1]
               ? ' / ' . i(a({href=>makeLink($i,$node,'r')}, $strings[1]))
               : '')
            . ($strings[2]
               ? ' / '.i(a({href=>makeLink($i,$node,'x')},$strings[2]))
               : '')
            . ' ';
    } elsif ($type eq 's') { #synset
        $DEBUG{'longHist'} and print " -> synset";
        my @gloss = splitGloss($node->gloss);
        $gloss[2] = $options{example} ? i($gloss[2]) : "";
        if ($options{count} && $node->parent->getobject eq 'over') {
            my $temp_count = $node->getSCcount;
            $histString .= "($temp_count)" if $temp_count;
        }
        $histString .= '{' . $node->getOffset . '} ' if $options{offset};
        $histString .= '&lt;' . $object->lexfilename . '&gt;' if $options{lex};
        $histString .= '['.$object->{lexfile}.'] ' if $options{lexnum};
        $histString .= a({href=>makeLink($i, $node)}, 'S:') . colorpos($node);
	my @wordStrings;
	for ($object->Words) {
		my $wordString = $_->word;
		$wordString .= '#'.$_->sensenumber if $options{sensenum};
		$wordString .= ' ('.$_->sk.')' if $options{sensekey};
		$wordString = (lc($_->word) eq lc($searchTerm))
			? b($wordString)
			: a( { href => makeLink('', '', '', $_->word) },
			     $wordString);
		push @wordStrings, $wordString;
	}
        $histString .= join ", ", @wordStrings;
        if ($options{gloss}) {
            $histString .= ' (' . ($gloss[0] ? '(' . $gloss[0] . ') ' : '')
                . $gloss[1] . ')';
        }
        $histString .= ' ' . $gloss[2];
        if ($i == $jump) { $histString = b($histString); }
    } else { #word
        if ($DEBUG{'longHist'}) { print " -> word", p; }
        my @prevWords = @{$node->relatedto};
        if ($DEBUG{'longHist'}) { print "; prevWords has "
                                      . scalar(@prevWords) . " items"; }
        my @gloss = splitGloss($node->gloss);
        $gloss[2] = $options{example} ? i($gloss[2]) : "";
        if ($DEBUG{'longHist'}) { print "; gloss=" . $gloss[1], p; }
        $histString = ($options{offset} ? '{' . $node->getOffset . '} ' : '')
            . ($options{lex} ? '&lt;'.$object->lexfilename.'&gt; ' : '')
            . a({href=>makeLink($i, $node)}, 'W:') . colorpos($node);
	my $wordString = $object->word;
	$wordString .= '#'. $object->sensenumber if $options{sensenum};
	$wordString .= ' ('.$object->sk.')' if $options{sensekey};
	$histString .= "<!-- ".($object->word)." =? $searchTerm [@prevWords] ($type) -->\n";
	$histString .= # (lc($object->word) eq lc($searchTerm))
		# ? b($wordString) # no self-links
#		:
 a({href=>makeLink('', '', '', $object->word)}, $wordString);
        if (@prevWords) {
            my $Rel = $node->parent;
            my $Par = $Rel->parent;
            if ($Par->type eq 's') { $Par = $Par->getobject; }
            else { $Par = $Par->getSynset; }
            $Rel = $Rel->getobject;
            if (grep { $prevWords[0]->id eq $_->id; } $Par->Words){
                if ($Rel eq 'ants') {
                    $histString .= " [Opposed to: ";
                } else {
                    $histString .= " [Related to: ";
                }
            } else {
                $histString .= " [Indirect via ";
            }
            $histString .= join(", ",
                map {$_ = a({href=>makeLink('', '', '', $_->word),
                             class=>'relWords'}, $_->word)}
                    @prevWords) . "]";
        }
        if ($options{gloss}) {
            $histString .= ' (' . ($gloss[0] ? '(' . $gloss[0] . ') ' : '')
                . $gloss[1] . ')';
        }
        $histString .= ' ' . $gloss[2];
        if ($i == $jump) { $histString = b($histString); }
    }
    return $histString;
}


sub colorpos {
    my $ps = $_[0]->pos;
    return a({class=>'pos'}, " (" . $shortPos{$ps} . ") ");
}


sub makeLink {
    my $info = shift;    # history index
    my $node = shift;    # node corresponding to $info
    my $opt = shift;     # option (for other types of links)
    my $newTerm = shift; # new searchTerm (if any)
    my $link = $self_url . '?' . ($LINKS_OPTS ? $LINKS_OPTS : '');
    if (defined($info) && !defined($opt) && !defined($newTerm)
        && @{$node->children} && $h[$info + 1] <= 1) {
        my @kids = $node->clip;
        my $temph = $history->geth;
        $node->children(@kids);
        $link .= 's=' . $searchTerm . '&h=' . $temph . '&j=' . $info . '#c';
        return escHTML($link);
    }
    if (defined($opt)) {
        if ($opt eq 'r') { $link .= 'r=1&' }          # recursive search
        elsif ($opt eq 'x') { $link .= 'r=2&' }       # 3rd kind of search
        elsif ($opt eq 'n') { return escHTML($link) } # new search
    }
    if ($newTerm) {
        return escHTML($link . 's=' . $newTerm);
    }
    if ($searchTerm) { $link .= 's=' . $searchTerm . '&' }
    if ($info ne '') { $link .= 'i=' . $info . '&' }
    if ($h ne "") { $link .= 'h=' . $h . '#c' }
    return escHTML($link);
}


sub escHTML {
    my $link = shift;
    $link =~ s/ /+/g;
    return $link;
}


sub splitGloss {
    my $gloss = shift;
    my @array = ();
    $gloss =~ m/^(\([^\)]*\) )?([^\"]*)(; \".*)?$/g;
    if (defined($1)) {
        push(@array, $1);
        $array[0] =~ s/^\(//g;
        $array[0] =~ s/\) $//g;
    } else { push(@array, ""); }
    if (defined($2)) { push(@array, $2); } else { push(@array, ""); }
    if (defined($3)) {
        push(@array, $3);
        $array[2] =~ s/^; //g;
    } else { push(@array, ""); }
    return @array;
}


sub printError {
	my $h3 = shift;
	$h3 = '' unless defined $h3;
	$h3 =~ s/([^A-Za-z0-9.;=+\-_\s\'()])/(?)/g;
    printForm();
    print h3($h3), end_html;
    exit 0;
}


sub fitBrowser {
# CSS style definitions
my $STYLE=<<END;
ul.typer {list-style-type: circle;}
ul.types {list-style-type: disc;}
ul.typew {list-style-type: disc;}
a.warning {color: red;}
a:link.relWords {color: green;}
a.pos {color: red;}
a.showAll {
  font-size: 75%;
  font-family: sans-serif; }
.title { font-size: 120%; }
.header { margin-bottom: .3em; }
.key { margin-top: .3em; }
END

my $PAD_STYLE=<<END;
ol,ul {padding-left: 1.4em;}
END

my $IE_STYLE=<<END;
line-height: 90%
END

    my $browser = user_agent()||'';
    if ($browser =~ m/[Nn]etscape/) {
        return $STYLE . $PAD_STYLE;
    } elsif ($browser =~ m/MSIE/) {
        return $STYLE . $IE_STYLE;
    } else {
        return $STYLE;
    }
}
